#from io import *
#from pjm import *
#from lupton import *
#from compose import *

#__all__ = ["io", "pjm", "lupton","compose"]


import numpy
import sys,getopt
from PIL import Image
import humvi
import os
import astropy.io.fits as pyfits



# =====================================================================

def compose(rfile, gfile, bfile, scales=(1.0,1.0,1.0), Q=1.0, alpha=1.0, \
            masklevel=None, saturation='color', offset=0.0, backsub=False, \
            vb=False, outfile='color.png'):
    """
    Compose RGB color image.
    """

    # -------------------------------------------------------------------

    if vb:
        print("HumVI: Making color composite image of data in following files:",rfile,gfile,bfile)
        print("HumVI: Output will be written to",outfile)
        if masklevel is not None: print ("HumVI: Masking stretched pixel values less than",masklevel)

    # Read in images, calibrated into flux units:

    band3 = humvi.channel(rfile)
    band2 = humvi.channel(gfile)
    band1 = humvi.channel(bfile)

    # Check shapes are equal:
    humvi.check_image_shapes(band1.image,band2.image,band3.image)

    # Subtract backgrounds (median, optional):
    if backsub:
      band1.subtract_background()
      band2.subtract_background()
      band3.subtract_background()

    # -------------------------------------------------------------------

    # BUG: as it stands, this code assumes one file one channel, whereas
    # in practice we might like to be able to make composites based on
    # N bands. Need to overload + operator for channels? Calib etc will
    # need altering as well as image.

    red = band3
    green = band2
    blue = band1

    # -------------------------------------------------------------------
    # Set scales determining color balance in composite:

    rscale,gscale,bscale = humvi.normalize_scales(scales)
    red.set_scale(manually=rscale)
    green.set_scale(manually=gscale)
    blue.set_scale(manually=bscale)
    if vb:print ('HumVI: Scales normalized to:',red.scale,green.scale,blue.scale)

    # Scale images - only do once:
    red.apply_scale()
    green.apply_scale()
    blue.apply_scale()

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Stretch images to cope with high dynamic range:

    if vb:
        print("HumVI: Stretch parameters Q,alpha:",Q,alpha)
        print("HumVI: At low surface brightness levels, the channel images are further rescaled by alpha")
        print("HumVI: Nonlinearity sets in at about 1/Q*alpha in the scaled intensity image:",1.0/(Q*alpha))

    # Compute total intensity image and the arcsinh of it:
    I = humvi.lupton_intensity(red.image,green.image,blue.image,type='sum')
    stretch = humvi.lupton_stretch(I,Q,alpha)

    # Apply stretch to channel images:
    r = stretch * red.image
    g = stretch * green.image
    b = stretch * blue.image

    if masklevel is not None:
        # Mask problem areas - exact zeros or very negative patches should
        # be set to zero.

        # BUG: this should have been done after scaling but before conversion
        # to channels, as its the individual images that have problems...

        r,g,b = humvi.pjm_mask(r,g,b,masklevel)

    # Offset the stretched images to make zero level appear dark gray.
    # Negative offset makes background more black...
    r,g,b = humvi.pjm_offset(r,g,b,offset)

    if saturation == 'color':
        # Saturate to colour at some level - might as well be 1, since
        # Q redefines scale?:
        threshold = 1.0
        r,g,b = humvi.lupton_saturate(r,g,b,threshold)
    # Otherwise, saturate to white.

    # Package into a python Image, and write out to file:
    image = humvi.pack_up(r,g,b)
    image.save(outfile)

    if vb: print( "HumVI: Image saved to:",outfile)

    return

# ======================================================================

vb = 0

# ======================================================================

# BUG: This code implies one file one channel, whereas we want to make
# composites based on N images... Class should be image, not channel.
# RGB channels should be constructed *after* scaling but *before* stretching

class channel:

    def __init__(self,fitsfile):

        self.input = fitsfile
        # Read in image and header:
        hdulist = pyfits.open(self.input)
        # self.hdr = hdulist[0].header
        # self.image = hdulist[0].data
        # Picking -1 header assumes we have 1 extension or PS1 (2 ext, image is last)
        self.image = hdulist[-1].data
        self.hdr = hdulist[-1].header
        self.calibrate()
        hdulist.close()

        return

    def calibrate(self):

        # Which telescope took these data?
        self.get_origin()

        # Get zero point, exptime:
        self.get_zeropoint()
        # EXPTIME is 1.0 for images in counts/s - but headers do not always
        # get this right... get_exptime gets this right.
        self.get_exptime()
        # Airmass? Gain? Should be included in zeropoint.

        # print "Image statistics for "+self.input
        # print "  ",self.origin,self.exptime,self.zpt

        # # Report 5 sigma depth:
        # image = self.image.copy()
        # mean = numpy.average(image)
        # stdev = numpy.std(image)
        # nsigma = 1
        # clip = 3
        # # Apply clipping:
        # while nsigma > 0.01:
        #     index = numpy.where(abs((self.image - mean)/stdev) < clip)[0]
        #     image = image[index]
        #     newmean = numpy.average(image)
        #     newstdev = numpy.std(image)
        #     nsigma = abs(mean - newmean)/newstdev
        #     mean = newmean
        #     stdev = newstdev
        # print "  Before calibration, mean, rms = ",mean,stdev
        # depth = -2.5*numpy.log10(5.0*stdev) + self.zpt
        # print "  Approximate 5-sigma limiting magnitude: ",depth

        # Compute calibration factor for image pixel values to
        # convert them into flux units. The 30 is arbitrary, and
        # simply determines the absolute value of alpha required
        # for a nice image.
        self.calib = (10.0**(0.4*(30.0 - self.zpt))) / self.exptime
        self.image *= self.calib

        # # Report 5 sigma depth:
        # image = self.image.copy()
        # mean = numpy.average(image)
        # stdev = numpy.std(image)
        # nsigma = 1
        # clip = 3
        # # Apply clipping:
        # while nsigma > 0.01:
        #     index = numpy.where(abs((self.image - mean)/stdev) < clip)[0]
        #     image = image[index]
        #     newmean = numpy.average(image)
        #     newstdev = numpy.std(image)
        #     nsigma = abs(mean - newmean)/newstdev
        #     mean = newmean
        #     stdev = newstdev
        # print "  After calibration, mean, rms = ",mean,stdev

        return

    def get_origin(self):
        if 'TELESCOP' in self.hdr:
            if self.hdr['TELESCOP'] == 'CFHT 3.6m':
                self.origin = 'CFHT'
            elif self.hdr['TELESCOP'] == 'ESO-VLT-U0':
                # Assume that all data from ESO-VLT-U0 is from KIDS.
                self.origin = "KIDS"
            else:
                self.origin = self.hdr['TELESCOP']
        elif 'ORIGIN' in self.hdr:
            if self.hdr['ORIGIN'] == 'CFHT':
                self.origin = 'CFHT'
            elif self.hdr['ORIGIN'] == 'DES':
                self.origin = 'DES'
            else:
                self.origin = self.hdr['ORIGIN']
        elif 'PSCAMERA' in self.hdr:
            self.origin = 'PS1'
        elif 'FID_ZP' in self.hdr:
            self.origin = 'DES'
        elif 'PROV' in self.hdr:
            self.origin = 'VICS82'
        else:
            self.origin = 'UNKNOWN'
        return

    def get_zeropoint(self):
        if self.origin == 'CFHT':
            if 'MZP_AB' in self.hdr:
                self.zpt = self.hdr['MZP_AB']
            elif 'MAGZP' in self.hdr:
                self.zpt = self.hdr['MAGZP']
            # elif 'PHOT_C' in self.hdr:
            #     self.zpt = self.hdr['PHOT_C']
            else:
                self.zpt = 30.0
        elif self.origin == 'PS1':
            self.zpt = self.hdr['HIERARCH FPA.ZP']
        elif self.origin == 'DES':
            self.zpt = self.hdr['MZP_AB']
        elif self.origin == 'VICS82':
            if 'MZP_AB' in self.hdr:
                self.zpt = self.hdr['MZP_AB']
            else:
                self.zpt = 30.0
        elif self.origin == 'KIDS':
            # KiDS coadds are calibrated to ZPT=0.
            self.zpt = 0.0
        else: # UNKNOWN
            self.zpt = 30.0
        return

    def get_exptime(self):
        # Here we assume that both CFHT and PS1 provide images with
        # pixel values in counts per second... or that the zero point
        # takes into account the exptime.
        if self.origin == 'CFHT':
            self.exptime = 1.0
        elif self.origin == 'PS1':
            # self.exptime = self.hdr['EXPTIME']
            self.exptime = 1.0
        elif self.origin == 'DES':
            # self.exptime = self.hdr['EXPTIME']
            self.exptime = 1.0
        elif self.origin == 'VICS82':
            # self.exptime = self.hdr['EXPTIME']
            self.exptime = 1.0
        elif self.origin == 'KIDS':
            # self.exptime = self.hdr['EXPTIME']
            self.exptime = 1.0
        else: #UNKNOWN
            # Use 1.0 as default to ensure the program doesn't crash.
            self.exptime = 1.0
        return

    def set_scale(self,manually=False):
        if manually:
             self.scale = manually
        else:
             self.scale = 1.0
        return

    def apply_scale(self):
        self.image *= self.scale
        return

    def subtract_background(self):
        self.image -= numpy.median(self.image)
        return

    def writefits(self):
        self.output = str.split(self.input,'.')[0]+'_calibrated.fits'
        if os.path.exists(self.output): os.remove(self.output)
        hdu = pyfits.PrimaryHDU()
        hdu.header = self.hdr
        hdu.data = self.image
        hdu.verify()
        hdu.writeto(self.output)
        return

# ======================================================================

def normalize_scales(scales):
    assert len(scales) == 3
    s1,s2,s3 = scales
    mean = (s1 + s2 + s3)/3.0
    return s1/mean, s2/mean, s3/mean

# ----------------------------------------------------------------------

def filter2wavelength(fname):

    # CFHT MegaCam (from http://www.cfht.hawaii.edu/Instruments/Imaging/Megacam/specsinformation.html)
    if fname == 'u.MP9301':
        L = 3740
    elif fname == 'g.MP9401':
        L = 4870
    elif fname == 'r.MP9601':
        L = 6250
    elif fname == 'i.MP9701' or 'i.MP9702':
        L = 7700
    elif fname == 'z.MP9801':
        L = 9000

    # SDSS:

    # DES:

    # etc
    return L

# ----------------------------------------------------------------------

def check_image_shapes(r,g,b):

    if (numpy.shape(r) != numpy.shape(g)) or \
        (numpy.shape(r) != numpy.shape(b)):
        raise "Image arrays are of different shapes, exiting"

    return

# ----------------------------------------------------------------------
# Make an 8 bit integer image cube from three channels:

def pack_up(r,g,b):

    NX,NY = numpy.shape(r)

    x = numpy.zeros([NX,NY,3])
    x[:,:,0] = numpy.flipud(r)
    x[:,:,1] = numpy.flipud(g)
    x[:,:,2] = numpy.flipud(b)

    x = numpy.clip(x,0.0,1.0)
    x = x*255

    return Image.fromarray(x.astype(numpy.uint8))

# ======================================================================


# ======================================================================

def lupton_intensity(r,g,b,type='sum'):

    if type == 'sum':
        return (r+g+b) + 1e-10

    elif type == 'rms':
        return numpy.sqrt(r*r+g*g+b*b) + 1e-10

# ----------------------------------------------------------------------

def lupton_stretch(I,Q,alpha):

    return numpy.arcsinh(alpha*Q*I) / (Q*I)

# ----------------------------------------------------------------------
# Clip high values to box:

def lupton_saturate(r,g,b,threshold):

    x = numpy.dstack((r,g,b))

    # Highest pixel-value at given position
    maxpix = numpy.max(x, axis=-1)
    maxpix[maxpix<1.0] = 1.0

    rr = r/maxpix
    gg = g/maxpix
    bb = b/maxpix

    return rr,gg,bb

# ======================================================================
# Testing:

if __name__ == '__main__':

    print("No tests defined")

# ======================================================================






# ======================================================================

"""
Functions for implementing Phil's tweaks to the Lupton algorithm for
making color composite images.
"""

# ======================================================================
# Globally useful modules:

import numpy
from PIL import Image

# ======================================================================
# Add small offset to image, to make background look dark gray not black:

def pjm_offset(r,g,b,offset):

    rr = r + offset
    gg = g + offset
    bb = b + offset

    return rr,gg,bb

# ----------------------------------------------------------------------
# Detect problem areas in any of the channel images, and mask out:

def pjm_mask(r,g,b,threshold):

    tiny = 1e-10
    mask = r*0.0 + 1.0

    for image in (r,g,b):

        image[numpy.isnan(image)] = 0.0
        image[numpy.isinf(image)] = 0.0

        mask[image < threshold] = 0.0
        mask[(image > -tiny) & (image < tiny)] = 0.0

    return r*mask,g*mask,b*mask

# ======================================================================




from astropy.io import ascii
import glob
import matplotlib.pyplot as plt
import numpy as np
from scipy import signal
from scipy.signal import medfilt
from astropy.io import fits
from astropy.table import Table
from scipy.constants import c
c_kms = c*1e-3

import logging
from scipy import interpolate
from IPython.display import Image, display
import os
import matplotlib.pyplot as plt
import matplotlib.image as mpimg
from IPython.display import display,Image
from matplotlib.patches import Rectangle
import humvi
import shutil

## Custom packages
from ISMgas.fitting.DoubleGaussian import *
from ISMgas.GalaxyProperties import *
from ISMgas.linelist import linelist_highz,linelist_SDSS
from ISMgas.SupportingFunctions import display_image


class kcwiAnalysis(GalaxyProperties):

    def __init__(self,**kwargs):

        GalaxyProperties.__init__(self,**kwargs)

        self.fileNames = kwargs.get('filename','')
        self.maskFile  = kwargs.get('maskfile','')
        self.varFile   = kwargs.get('varfile','')


        self.combine   = kwargs.get('combine','mean')

        if(self.combine=='mean'):
            self.dataCube = self.return_meancube()

        elif(self.combine=='median'):
            self.dataCube = self.return_mediancube()
            
        if(self.maskFile==''):
            self.maskData = np.array([])
        else:
            self.maskData = fits.getdata(self.maskFile)
        
        if(self.varFile==''):
            self.varData  = np.array([])
            self.errData  = np.array([])
        else:
            self.varData  = fits.getdata(self.varFile)
            self.errData  = np.sqrt(self.varData)


        self.meanStart    = kwargs.get('meanstart',500)
        self.meanEnd     = kwargs.get('meanend', 1500)

        self.dataCubeMean = np.mean(self.dataCube[self.meanStart:self.meanEnd,:,:],0)

        self.rscale, self.gscale, self.bscale     = kwargs.get('scale', [1,1.4,2])
        self.Q, self.alpha                        = kwargs.get('q_alpha', [3,0.4])
        self.masklevel, self.maskoffset           = kwargs.get('mask_offset', [-1.0,0.0])
        self.saturation, self.backsub, self.vb    = kwargs.get('saturation_backsub_vb',['white',False,False])


    def whiteLightImage(self,**kwargs):
        plt.imshow(
            self.dataCubeMean,
            origin    = 'lower',
            cmap      = kwargs.get('cmap'),
            vmin      = kwargs.get('vmin'),
            vmax      = kwargs.get('vmax')
        )

    def overlayMask(self,**kwargs):
        plt.imshow(
            self.maskData,
            origin    = 'lower',
            cmap      = kwargs.get('cmap'),
            vmin      = kwargs.get('vmin'),
            vmax      = kwargs.get('vmax'),
            alpha     = kwargs.get('alpha',0.2)
        )


    def return_mediancube(self):
        data_temp = []
        for i in self.fileNames:
            data_temp.append(fits.getdata(i))

        return(np.median(np.array(data_temp),axis=0))

    def return_meancube(self):
        data_temp = []
        for i in self.fileNames:
            data_temp.append(fits.getdata(i))
        return(np.mean(np.array(data_temp),axis=0))


    def humviPNG(self, bwave=[500,1000], gwave=[800,1200], rwave=[1200,1500]):
        """
        This function will create a png file using the humvi package.
        rwave, gwave, bwave are the wavelength ranges for the red, green, and blue channels.
        """
        bfile   = np.sum(self.dataCube[bwave[0]:bwave[1]],0)
        hdu     = fits.PrimaryHDU(data=bfile)
        hdu.writeto(f"{self.objid}_B.fits",overwrite=True)

        gfile   = np.sum(self.dataCube[gwave[0]:gwave[1]],0)
        hdu     = fits.PrimaryHDU(data=gfile)
        hdu.writeto(f"{self.objid}_G.fits",overwrite=True)

        rfile   = np.sum(self.dataCube[rwave[0]:rwave[1]],0)
        hdu     = fits.PrimaryHDU(data=rfile)
        hdu.writeto(f"{self.objid}_R.fits",overwrite=True)
        
        humvi.compose(
            f"{self.objid}_R.fits",
            f"{self.objid}_G.fits",
            f"{self.objid}_B.fits",
            scales     = (self.rscale, self.gscale, self.bscale),
            Q          = self.Q,
            alpha      = self.alpha,
            masklevel  = self.masklevel,
            saturation = self.saturation,
            offset     = self.maskoffset,
            backsub    = self.backsub,
            vb         = self.vb,
            outfile    = self.objid+"_cubePng.png"
        )

        display_image(self.objid+"_cubePng.png")



import numpy as np
import matplotlib.pyplot as plt
from matplotlib import gridspec

from symfit import variables, parameters, Model, Fit, exp, GreaterThan, LessThan
from symfit.core.minimizers import BasinHopping

import time
from ISMgas.fitting.init_fitting import *
from ISMgas.SupportingFunctions import load_pickle, save_as_pickle, plotHistogram
from ISMgas.globalVars import *

class DoubleGaussian:
    '''
    Class which performs the double / single gaussian fit to the data.
    Assumes that continuum is at a constant level. Preprocess data if needed   
    '''
    def __init__(self, x, y, yerr, inst_sigma=1):
        '''
        x, y, yerr : required 
        inst_sigma : (optional) minimum sigma that the spectra can resolve. Essential for derving velocity measurements
        '''
       
        self.x            = x
        self.y            = y
        self.yerr         = yerr
        self.inst_sigma   = inst_sigma
        self.init_values  = {}
        self.results_dict = {}
        
        # use priors in init_fitting.py or make your own priors on the go!
        self.priors   = priorsInit()


    def find_nearest(self, input_array, value):
        '''
        used in derivedVelocityMeasurements. 
        Given an array and a value X, returns the value in array nearest to X.
        '''
        array   = np.asarray(input_array)
        idx     = (np.abs(array - value)).argmin()
        return array[idx]


    def derivedVelocityMeasurements(self, A_out, A1_out, v_out, v1_out, sig_out, sig1_out, double_gaussian, outflow=False):
        '''
            This function returns all the derived velocity measurements given the double gaussian coefficients

            - v01
            - v05
            - v10
            - v25
            - v50
            - v75
            - v90
            - v95
            - v99,
            - $\Delta$ v98
            - $\Delta$ v90
            - $\Delta$ v80
            - $\Delta$ v50

            #### Example

            ```
            obj.derivedVelocityMeasurements(1,2,3,4,5,6,double_gaussian=True,outflow=False)

            {'v01': -10,
             'v05': -6,
             'v10': -4,
             'v25': -1,
             'v50': 3,
             'v75': 7,
             'v90': 11,
             'v95': 13,
             'v99': 17,
             'delta_v98': 27,
             'delta_v90': 19,
             'delta_v80': 15,
             'delta_v50': 8}

            ```

        '''
        v = np.arange(-1300,1300,1)
        if(outflow):
            v = np.arange(-1300,0,1)

        if(double_gaussian):
            ## NOTE: This uses (A*(exp(-(w**2)/2.))+  A1*(exp(-(w1**2)/2.)) whereas the fitting uses 1-[..]

            flux = np.array([ A_out*np.exp(-0.5*(v_out -i)**2/sig_out**2) +A1_out*np.exp(-0.5*(v1_out - i)**2/sig1_out**2) for i in v])

        elif(double_gaussian==False):

            flux = np.array([ A_out*np.exp(-0.5*(v_out -i)**2/sig_out**2)  for i in v])

        cdf = np.cumsum(flux)/np.max(np.cumsum(flux)) 

        ## Alternate implementation could be using np.interp but since we already know the functional form this is not necessary
        ## Due to the legacy reasons v05 here represents 95th percentile of absorption and v25 the 75th percentile of absorption 
        ## defined in the conventional sense. 
               
        measurements = {
            'v01': v[np.where(cdf==self.find_nearest(cdf,0.01))][0],
            'v05': v[np.where(cdf==self.find_nearest(cdf,0.05))][0],
            'v10': v[np.where(cdf==self.find_nearest(cdf,0.10))][0],
            'v25': v[np.where(cdf==self.find_nearest(cdf,0.25))][0],
            'v50': v[np.where(cdf==self.find_nearest(cdf,0.50))][0],
            'v75': v[np.where(cdf==self.find_nearest(cdf,0.75))][0],
            'v90': v[np.where(cdf==self.find_nearest(cdf,0.90))][0],
            'v95': v[np.where(cdf==self.find_nearest(cdf,0.95))][0],
            'v99': v[np.where(cdf==self.find_nearest(cdf,0.99))][0]
        }

        measurements['delta_v98']   = measurements['v99'] -  measurements['v01']
        measurements['delta_v90']   = measurements['v95'] -  measurements['v05']
        measurements['delta_v80']   = measurements['v90'] -  measurements['v10']
        measurements['delta_v50']   = measurements['v75'] -  measurements['v25']
        
        measurements['EW_kms']      = np.sum(flux) ## EW_kms = \sum C_f(v) * 1 (km/s) -- we are sampling it at 1 km/s

        return(measurements)


    def fitting(self,
                niter               = 100,
                seed                = 12345,
                constraint_option   = 1,
                double_gaussian     = True,
                verbose             = False,
                continuum           = [],
                stepsize            = 0.005
        ):
        '''
        Given x, y and ysigma the function returns the parameters of a
        double gaussian fit of the form :
        1- (A*(exp(-(w**2)/2.))-  A1*(exp(-(w1**2)/2.))

        or a single gaussian fit of the form
        1- (A*(exp(-(w**2)/2.))


        #### Example
        ```
        A = 0.8
        A1 = 0.6
        sig = 220
        sig1 = 150
        v = -250
        v1 = 10


        obj.combined_wavelength = np.arange(-2000,2000,25)
        obj.combined_flux = 1-(
            A*np.exp(-(v-obj.combined_wavelength)**2/(sig**2)*2.)
            +A1*np.exp(-(v1-obj.combined_wavelength)**2/(sig1**2)*2.))


        plt.figure(figsize=(12,8))
        plt.plot(obj.combined_wavelength,obj.combined_flux )


        obj.combined_fluxerr = 0.05*np.ones(len(obj.combined_wavelength))
        results = obj.fitting(seed=12345,double_gaussian=True,verbose=False)

        plt.plot(results['fitted_wav'],results['fitted_flux'])

        ```

        '''

        wav_copy            = np.array(self.x.copy())
        final_profile_copy  = np.array(self.y.copy())
        final_error_copy    = np.array(self.yerr.copy())

        chi_sq_continuum = 0.
        if(len(continuum)>0):
            for i in continuum:
                continuum_profile   = final_profile_copy[(wav_copy>i[0]) & (wav_copy<i[1])]
                continuum_error     = final_error_copy[(wav_copy>i[0]) & (wav_copy<i[1])]
                chi_sq_continuum    += np.sum(((continuum_profile-1)**2)/continuum_error**2,axis=0)


        ##############################################
        ######### Priors and initialization ##########
        ##############################################

        x, y                      = variables('x,y')                      # x- velocity , y - flux
        v, sig, v1, sig1, A, A1   = parameters('v, sig, v1, sig1, A, A1') # parameters to optimize
        w                         = (v-x)/sig
        w1                        = (v1-x)/sig1
        
        ## Pick a random every single time the fitting is run ##
        np.random.seed(int(time.time())+seed)

        xdata = wav_copy

        ## At each point, generate a perturbed realization to fit by adding 1sigma noise
        ydata = final_profile_copy + np.array([np.random.uniform(low=-i,high=i) for i in final_error_copy])

        if(double_gaussian==True):
            
            A.value                     = np.random.uniform(low=self.priors[str(constraint_option)]["Amin"],high=self.priors[str(constraint_option)]["Amax"])
            self.init_values['A']       = A.value
            A.min                       = self.priors[str(constraint_option)]["Amin"]
            A.max                       = self.priors[str(constraint_option)]["Amax"]

            v.value                     = np.random.uniform(low=self.priors[str(constraint_option)]["vmin"],high=self.priors[str(constraint_option)]["vmax"])
            self.init_values['v']       = v.value
            v.min                       = self.priors[str(constraint_option)]["vmin"]
            v.max                       = self.priors[str(constraint_option)]["vmax"]

            sig.value                   = np.random.uniform(low=self.priors[str(constraint_option)]["sigmin"],high=self.priors[str(constraint_option)]["sigmax"])
            self.init_values['sig']     = sig.value
            sig.min                     = self.priors[str(constraint_option)]["sigmin"]
            sig.max                     = self.priors[str(constraint_option)]["sigmax"]

            A1.value                    = np.random.uniform(low=self.priors[str(constraint_option)]["A1min"],high=self.priors[str(constraint_option)]["A1max"])
            self.init_values['A1']      = A1.value
            A1.min                      = self.priors[str(constraint_option)]["A1min"]
            A1.max                      = self.priors[str(constraint_option)]["A1max"]

            v1.value                    = np.random.uniform(low=self.priors[str(constraint_option)]["v1min"],high=self.priors[str(constraint_option)]["v1max"])
            self.init_values['v1']      = v1.value
            v1.min                      = self.priors[str(constraint_option)]["v1min"]
            v1.max                      = self.priors[str(constraint_option)]["v1max"]

            sig1.value                  = np.random.uniform(low=self.priors[str(constraint_option)]["sig1min"],high=self.priors[str(constraint_option)]["sig1max"])
            self.init_values['sig1']    = sig1.value
            sig1.min                    = self.priors[str(constraint_option)]["sig1min"]
            sig1.max                    = self.priors[str(constraint_option)]["sig1max"]           
            
            model_dict = {
                y: self.priors[str(constraint_option)]["cont_lvl"] - A*(exp(-(w**2)/2.))-  A1*(exp(-(w1**2)/2.)).as_expr()
            }

            model = Model(model_dict)
            if(verbose):
                print(model)

            constraints = [
                LessThan(v-v1,self.priors[str(constraint_option)]["v-v1_min"]),
                GreaterThan(v-v1,self.priors[str(constraint_option)]["v-v1_max"])
            ]

            # Perform the fit - sigma_y is needed for chi^2 optimization
            fit = Fit(
                model,
                x             = xdata,
                y             = ydata,
                sigma_y       = final_error_copy,
                minimizer     = BasinHopping,
                constraints   = constraints
            )

        elif(double_gaussian==False):
            
            A.value                     = np.random.uniform(low=self.priors[str(constraint_option)]["Amin_SG"],high=self.priors[str(constraint_option)]["Amax_SG"])
            self.init_values['A']       = A.value
            self.init_values['A1']      = 0
            A.min                       = self.priors[str(constraint_option)]["Amin_SG"]
            A.max                       = self.priors[str(constraint_option)]["Amax_SG"]


            v.value                     = np.random.uniform(low=self.priors[str(constraint_option)]["vmin_SG"],high=self.priors[str(constraint_option)]["vmax_SG"])
            self.init_values['v']       = v.value
            self.init_values['v1']      = 0
            v.min                       = self.priors[str(constraint_option)]["vmin_SG"]
            v.max                       = self.priors[str(constraint_option)]["vmax_SG"]


            sig.value                   = np.random.uniform(low=self.priors[str(constraint_option)]["sigmin_SG"],high=self.priors[str(constraint_option)]["sigmax_SG"])
            self.init_values['sig']     = sig.value
            self.init_values['sig1']    = 0
            sig.min                     = self.priors[str(constraint_option)]["sigmin_SG"]
            sig.max                     = self.priors[str(constraint_option)]["sigmax_SG"]
           
            
            model_dict = {
                y: self.priors[str(constraint_option)]["cont_lvl"] - A*(exp(-(w**2)/2.)).as_expr()
            }

            model = Model(model_dict)
            if(verbose):
                print(model)

            # Perform the fit - sigma_y is needed for chi^2 optimization
            fit = Fit(
                model,
                x         = xdata,
                y         = ydata,
                sigma_y   = final_error_copy,
                minimizer = BasinHopping
            )


        fit_result = fit.execute(seed=int(time.time())+seed,stepsize=stepsize)
        # zfit = model(x = xdata, **fit_result.params)

        if(verbose):
            print(fit_result)

        if(double_gaussian==True):
            A_out         = fit_result.params['A']
            A_out_sig     = fit_result.stdev(A)

            A1_out        = fit_result.params['A1']
            A1_out_sig    = fit_result.stdev(A1)

            sig_out       = fit_result.params['sig']
            sig_out_sig   = fit_result.stdev(sig)

            sig1_out      = fit_result.params['sig1']
            sig1_out_sig  = fit_result.stdev(sig1)

            v_out         = fit_result.params['v']
            v_out_sig     = fit_result.stdev(v)

            v1_out        = fit_result.params['v1']
            v1_out_sig    = fit_result.stdev(v1)

            ### Deconvolved absorption profile values
            if(self.inst_sigma>sig_out or self.inst_sigma>sig1_out):
                raise ValueError("Chosen instrument resolution will be unable to resolve the ISM line. Please check the chosen instrument sigma (km/s)")

            sig_out_deconv  = np.sqrt(sig_out**2 - self.inst_sigma**2)
            sig1_out_deconv = np.sqrt(sig1_out**2 - self.inst_sigma**2)

            A_out_deconv    = A_out*(sig_out/sig_out_deconv)
            A1_out_deconv   = A1_out*(sig1_out/sig1_out_deconv)

            v_out_deconv    = v_out
            v1_out_deconv   = v1_out

            ## double gaussian function
            y00 = [ -A_out*np.exp(-0.5*(v_out -i)**2/sig_out**2) - A1_out*np.exp(-0.5*(v1_out - i)**2/sig1_out**2) + self.priors[str(constraint_option)]["cont_lvl"] for i in wav_copy]
            ## First component
            y01 = [ -A_out*np.exp(-0.5*(v_out -i)**2/sig_out**2) + self.priors[str(constraint_option)]["cont_lvl"] for i in wav_copy ]
            ## Second component
            y02 = [ -A1_out*np.exp(-0.5*(v1_out - i)**2/sig1_out**2) + self.priors[str(constraint_option)]["cont_lvl"] for i in wav_copy]

            chi_sq = np.sum(((np.array(y00)-final_profile_copy)**2)/final_error_copy**2,axis=0)


        elif(double_gaussian==False):
            A_out        = fit_result.params['A']
            A_out_sig    = fit_result.stdev(A)

            A1_out       = 0
            A1_out_sig   = 0

            sig_out      = fit_result.params['sig']
            sig_out_sig  = fit_result.stdev(sig)

            sig1_out     = 0
            sig1_out_sig = 0

            v_out        = fit_result.params['v']
            v_out_sig    = fit_result.stdev(v)

            v1_out       = 0
            v1_out_sig   = 0

            ### Deconvolved absorption profile values
            if(self.inst_sigma>sig_out):
                raise ValueError("Chosen instrument resolution will be unable to resolve the ISM line. Please check the chosen instrument sigma (km/s)")


            sig_out_deconv  = np.sqrt(sig_out**2 - self.inst_sigma**2)
            sig1_out_deconv = 0

            A_out_deconv    = A_out*(sig_out/sig_out_deconv)
            A1_out_deconv   = 0

            v_out_deconv    = v_out
            v1_out_deconv   = 0


            ## double gaussian function
            y00 = [ -A_out*np.exp(-0.5*(v_out -i)**2/sig_out**2) + self.priors[str(constraint_option)]["cont_lvl"] for i in wav_copy]
            ## First component
            y01 = [ -A_out*np.exp(-0.5*(v_out -i)**2/sig_out**2) + self.priors[str(constraint_option)]["cont_lvl"] for i in wav_copy]
            ## Second component
            y02 = [ self.priors[str(constraint_option)]["cont_lvl"] for i in xdata]

            chi_sq = np.sum(((np.array(y00)-final_profile_copy)**2)/final_error_copy**2,axis=0)

        ### Pack all results
        self.results_dict = {
            'niter'              : niter,
            'seed'               : seed,
            'cont_lvl'           : self.priors[str(constraint_option)]["cont_lvl"], ## Continuum level
            'constraint_option'  : constraint_option,                               ## which constraint option was used
            'fit_priors_used'    : self.priors[str(constraint_option)],             ## Copy of all the priors used.
            'A_out'              : A_out,
            'A_out_sig'          : A_out_sig,
            'A1_out'             : A1_out,
            'A1_out_sig'         : A1_out_sig,
            'sig_out'            : sig_out,
            'sig_out_sig'        : sig_out_sig,
            'sig1_out'           : sig1_out,
            'sig1_out_sig'       : sig1_out_sig,
            'v_out'              : v_out,
            'v_out_sig'          : v_out_sig,
            'v1_out'             : v1_out,
            'v1_out_sig'         : v1_out_sig,
            'chi_sq'             : chi_sq - chi_sq_continuum,
            'chi_sq_cont'        : chi_sq_continuum,
            
            'fitted_wav'         : wav_copy,
            'fitted_flux'        : y00,
            'fitted_flux_comp1'  : y01,
            'fitted_flux_comp2'  : y02,

            'A_out_deconv'       : A_out_deconv,
            'A1_out_deconv'      : A1_out_deconv,
            'sig_out_deconv'     : sig_out_deconv,
            'sig1_out_deconv'    : sig1_out_deconv,
            'v_out_deconv'       : v_out_deconv,
            'v1_out_deconv'      : v1_out_deconv,


        }

        self.results_dict['residual']          = np.sum((np.array(y00) - final_profile_copy)**2,axis=0)

        self.results_dict['initial-values']    = self.init_values

        self.results_dict['derived_results']   = self.derivedVelocityMeasurements(
            A_out,
            A1_out,
            v_out,
            v1_out,
            sig_out,
            sig1_out,
            double_gaussian
        )

        self.results_dict['derived_results_deconv'] = self.derivedVelocityMeasurements(
            A_out_deconv,
            A1_out_deconv,
            v_out_deconv,
            v1_out_deconv,
            sig_out_deconv,
            sig1_out_deconv,
            double_gaussian
        )


        self.results_dict['derived_results_onlyoutflow'] = self.derivedVelocityMeasurements(
            A_out,
            A1_out,
            v_out,
            v1_out,
            sig_out,
            sig1_out,
            double_gaussian,
            outflow=True
        )

        return(self.results_dict)


'''
Class handling i/o of spectra and making a velocity profile.
This code has been rigourously tested with data obtained from ESI and MagE. 
Feel free to try it out with other instruments. If you run into any issues, 
please report it on the repo                                                
'''

## Packages for plotting
import matplotlib.pyplot as plt
from matplotlib.pyplot import cm

## Astropy
from astropy.io import fits
from astropy.table import Table

## Custom packages
from ISMgas.fitting.DoubleGaussian import *
from ISMgas.GalaxyProperties import *
from ISMgas.linelist import linelist_highz
from ISMgas.globalVars import *
from ISMgas.SupportingFunctions import beautifyPlot
 

class AnalyzeSpectra(GalaxyProperties):
    '''
    Class handing the analysis of 1D spectra.  
    Example usage :
    ```
    obj = AnalyzeSpectra(
        objid           = "stack",
        spec_filename   = "stack.fits",
        R               = 4000,
        zs              = 0
    )


    obj.combineISMLines(
        Nsm             = 1,
        chosen_lines    = ['Si II 1260', 'Si II 1526'],
        wavrange        = [
            [-2000,2000],
            [-2000,2000]
            
        ],
    )
    foo = load_pickle('stack-combined.pkls')
    plotWithError(
        foo['source_wav'],
        foo['source_flux'],
        foo['source_error']
    )

    plt.ylim([0,1.3])


    ```
    '''

    def __init__(self, **kwargs):

        GalaxyProperties.__init__(self,**kwargs)
        
        self.color_pal                    = color_pal_kvgc['pub1']
        self.combined_wavelength          = []
        self.combined_flux                = []
        self.combined_fluxerr             = []
        self.all_lines                    = []
        self.all_lines_err                = []
        self.line_names                   = []
        self.user_wavrange                = []

        self.user_chosen_lines            = []
        self.user_chosen_lines_err        = []
        self.user_chosen_lines_name       = []
        self.user_chosen_lines_wavrange   = []

        self.Nsm                          = 0
        self.deltaWavelength              = 25 ## in \AA

    def fetchData(self):
        hdu   = fits.open(self.spec_filename)
        t     = Table(hdu[1].data)
        wl    = np.array(t['LAMBDA'])[0]
        spec  = np.array(t['SPEC'])[0]
        ivar  = np.array(t['IVAR'])[0]
        
        return(wl, spec, ivar)
      
    def dataloader(
            self,
            Nsm,
            state,
            wavrange           = [],
            plotting           = False,
            chosen_lines       = 0,
            start_interp       = -2000,
            end_interp         = 2000,
            delta_interp       = 25,
            xlim               = [-2000,2000],
            ylim               = [0,2],
            interpolate_spec   = True 
        ):
        
        '''
        Handles importing data from pre-processed fits file. 
        returns (wavelength, flux, sigma, line_name)
        '''

        ### Make a dedicated user linelist or use linelist.py 
        if chosen_lines == 0:
            chosen_lines = linelist_highz.keys()

        count             = 1
        wav               = []
        spectra_interp    = []
        line_name         = []
        sigma             = []     

        if plotting:
            plt.figure(figsize = [13,18])

        for q in chosen_lines:
            linename   = q
            line_wl    = linelist_highz[q]["lambda"]

            if linelist_highz[q].get(state) is not None:
                wl, spec, ivar = self.fetchData()

                ## Choosing wavelengths +-deltaWavelength A from the absorption line 
                start   = (1+self.zs)*(line_wl - self.deltaWavelength)
                end     = (1+self.zs)*(line_wl + self.deltaWavelength)

                mask_start    = wl > start
                mask_end      = wl < end
                mask_wl       = mask_start & mask_end
                
                wavelength      = wl[mask_wl]
                flux            = spec[mask_wl]
                invvar          = ivar[mask_wl]             
     
                wavelength_rest = wavelength/(1+self.zs)

                ## Sanity check to ensure that the flux array has elements in it
                if(len(flux)!=0):
                    
                    ## Take weighted mean of Nsm elements
                    f_wtd, ww_wtd, sigma_wtd = returnWeightedArray(
                        Nsm          = Nsm,
                        spec         = flux,
                        ivar         = invvar,
                        wave_rest    = c_kms*(wavelength_rest-line_wl)/line_wl  ## changing from angstrom to km/s
                    )
                    
                    #*********************************************#
                    if(plotting):
                        plt.subplot(8,4,count)
                        plt.plot(
                            ww_wtd, f_wtd,
                            color=self.color_pal[14],
                            alpha = 0.8,
                            linewidth=3,
                            label='%s-%d'%(linename,line_wl)
                        )

                        if(len(wavrange)>0):
                            plt.axvline(
                                [wavrange[count-1][0]],
                                color='black',
                                linewidth=2,
                                linestyle='--'
                            )
                            plt.axvline(
                                [wavrange[count-1][1]],
                                color='black',
                                linewidth=2,
                                linestyle='--'
                            )

                        elif(len(wavrange)==0):
                            plt.text(
                                1300,0.3,
                                count,
                                fontsize=50,
                                alpha=0.4,
                                horizontalalignment='center',
                                verticalalignment='center'
                            )

                        font_size=15
                        print("setting title to " + linename)
                        plt.title(linename,fontsize=font_size)
                        plt.xlabel('Velocity(km/s)',fontsize=font_size)
                        plt.ylabel('Flux',fontsize=font_size)
                        plt.xlim(xlim)
                        plt.ylim(ylim)
                        plt.xticks(np.arange(xlim[0],xlim[1]+1,1000),fontsize=12)
                        plt.yticks(np.arange(ylim[0],ylim[1],0.5),fontsize=12)
                        plt.grid('on',alpha=0.6)
                        plt.tight_layout()
                   #*********************************************#

                    count+=1

                    ## Begin Interpolation ##
                    if(interpolate_spec):
                        wav_interp    = np.arange(start_interp, end_interp, delta_interp)
                        wav.append(wav_interp)
                        
                        spectra_interp.append(
                            interpolateData(
                                    x       = ww_wtd,
                                    y       = f_wtd,
                                    xnew    = wav_interp
                            )
                        )
                        sigma.append(
                            interpolateData(
                                x       = ww_wtd,
                                y       = sigma_wtd,
                                xnew    = wav_interp
                            )
                        )
                        line_name.append(linename)
                    
                    else:
                        wav.append(ww_wtd)
                        spectra_interp.append(f_wtd)
                        sigma.append(sigma_wtd)
                        
                    ## End interpolation ##

        if plotting:
            fooPlotDict = {
                'lowion'    : {'title':"Low ionization ISM Lines", "saveFile": "%s-all-lines-lowions.png"%(self.objid)},
                'highion'   : {'title':"High ionization ISM Lines", "saveFile": "%s-all-lines-highions.png"%(self.objid)},
                'stellar'   : {'title':"Stellar Lines", "saveFile": "%s-stellar-lines.png"%(self.objid)},
                'opthin'    : {'title':"Optically thin Lines", "saveFile": "%s-opthin-lines.png"%(self.objid)},
                'fineem'    : {'title':"Fine structure Lines", "saveFile": "%s-fineem-lines.png"%(self.objid)},
                'nebem'     : {'title':"Nebular emission Lines", "saveFile": "%s-nebem-lines.png"%(self.objid)}
            }
            plt.suptitle(fooPlotDict[state]['title'],fontsize =25, y = 1.05)
            plt.savefig(fooPlotDict[state]['saveFile'],**ISMgasPlot['savefig'])

            plt.close()     
                           
        return(np.array(wav), np.array(spectra_interp), np.array(sigma), line_name)

    def plotLines(self, Nsm, xlim=[-2000,2000], ylim=[0,2], state='lowion'):
        '''
        Plots the individual ISM lines from spectra. Plots the raw spectra without interpolation.
        '''

        self.dataloader (
            Nsm                 = Nsm,
            state               = state,
            xlim                = xlim,
            ylim                = ylim,
            plotting            = True,
            interpolate_spec    = False,
        )

    def combineISMLines(
            self,
            Nsm,
            chosen_lines,
            wavrange        = [],
            xlim            = [-2000,2000],
            ylim            = [0,2],
            start_interp    = -2000,
            end_interp      = 2000,
            delta_interp    = 25,
            state           = 'lowion',
            normScaleFactor = 1,
            roi             = [] ):
        '''
            returns a weighted average of all the lines chosen by the user
        '''

        wav, spectra_interp, sigma, line_name = self.dataloader(
            Nsm            = Nsm,
            chosen_lines   = chosen_lines,
            start_interp   = start_interp,
            end_interp     = end_interp,
            delta_interp   = delta_interp,
            wavrange       = wavrange,
            state          = state,
            plotting       = True
        )
   
        # Normalize all spectra using median (RUN THIS ONLY ONCE!)
        for i in reversed(range(len(spectra_interp))):
            median_val = np.median(spectra_interp[i])

            # Multiply by normscalefactor (Default normScaleFactor=1) so this step does not do anything
            # But this is useful when the flux is not normalized to 1 but needs manual tweaking...
            median_val *= normScaleFactor

            # Divide both the spectra and the sigma by this median value
            spectra_interp[i]   = spectra_interp[i]/median_val
            sigma[i]            = sigma[i]/median_val
        
        ##-- Save --##
        allLinesSpec = {
            'wav'               : wav,
            'spectra_interp'    : spectra_interp,
            'sigma'             : sigma,
            'line_name'         : line_name,
            'wavrange'          : wavrange
        }
        save_as_pickle(allLinesSpec, fileName = f"{self.objid}-allLinesSpec.pkls")
        self.plotVerticalStack()
        
        if(len(wavrange)==0):
            print("Provide wavelength ranges and rerun ")
            
        else:
            self.user_wavrange    = wavrange
            self.Nsm              = Nsm
            
            ## Create inverse variance array, from standard deviation
            ## Note that the above could give problems if there are any cases of sigma=0
            invvar = []
            for i in range(len(spectra_interp)):
                invvar.append(sigma[i]**(-2.))
                
            ## TJ: set inverse-variance to 0 to mask out regions outside the wavrange.
            for i in range(len(spectra_interp)):
                invvar[i][ wav[i] < wavrange[i][0] ]  = 0
                invvar[i][ wav[i] > wavrange[i][1] ]  = 0

            ## TJ version of combined average profile :  We have X lines with each line having an ivar. 
            ## Take the inversevariance-weighted mean of all the X lines to get the final combined absorption profile
            f_interp        = np.array(spectra_interp)
            ivar_interp     = np.array(invvar)

            # Inverse-variance weighted mean of the absorption lines
            final_profile   = np.sum(f_interp*ivar_interp, 0) / np.sum(ivar_interp, 0)
            final_ivar      = np.sum(ivar_interp, 0)
            final_error     = 1./np.sqrt(final_ivar)

            # Fitting final profile continuum with polynomial
            if(len(roi)>0):

                mask_continuum0       = wav[0] < roi[0]
                mask_continuum1       = wav[0] > roi[1]
                mask_continuum        = mask_continuum0 | mask_continuum1

                continuum_fit         = np.polyfit(
                    wav[0][mask_continuum],
                    final_profile[mask_continuum],
                    3
                )
                continuum_fit_poly    = np.poly1d(continuum_fit)

                final_profile   = final_profile/continuum_fit_poly(wav[0])
                final_error     = final_error/continuum_fit_poly(wav[0])
                
                continuumFitVals = {
                    'wav'                   : wav[0],
                    'final_profile'         : final_profile,
                    'final_error'           : final_error,
                    'continuum_fit_poly'    : continuum_fit_poly(wav[0])
                }
                save_as_pickle(continuumFitVals, fileName=f"{self.objid}-continuumfit.pkls")
                self.plotContinuumFit()

            #*********************************************#

            self.combined_wavelength    = wav[0]
            self.combined_flux          = final_profile
            self.combined_fluxerr       = final_error
            self.all_lines              = f_interp
            self.all_lines_err          = 1./np.sqrt(ivar_interp)
            self.line_names             = line_name


            computedSpecVals = {
                'source_wav'        : self.combined_wavelength,
                'source_flux'       : self.combined_flux,
                'source_error'      : self.combined_fluxerr,
                'wav'               : wav,
                'spectra_interp'    : f_interp,
                'sigma'             : 1./np.sqrt(ivar_interp),
                'line_name'         : line_name,    
                'wavrange'          : wavrange,           
                'state'             : state
            }
            save_as_pickle(computedSpecVals, fileName = f"{self.objid}-combined.pkls")
            self.plotCombinedLines()
            
            

            
    ##########################
    ### Plotting routines  ###
    ##########################
    
    def plotVerticalStack(self, xlim=[-2000,2000], ylim=[0,2]):
        fooDict = load_pickle(fileName = f"{self.objid}-allLinesSpec.pkls")
        
        spectra_interp    = fooDict['spectra_interp']
        line_name         = fooDict['line_name']
        wav               = fooDict['wav']

        plt.figure(figsize = [7,8])
        
        count = len(self.color_pal)-1

        for i in reversed(range(len(spectra_interp))):
            plt.plot(    
                wav[i],
                spectra_interp[i]+i,
                alpha       = 0.8,
                linewidth   = 3,
                label       = line_name[i],
                color       = self.color_pal[count]
            )
            count = count-1
            
            beautifyPlot({                     
                'xlabel'  : {'xlabel': 'Velocity(km/s)'},
                'ylabel'  : {'ylabel': 'Flux (Normalized)'},

                'xlim'    : {'left': xlim[0],'right': xlim[1]},
                'ylim'    : {'bottom': ylim[0],'top': None},

                
                'tightlayout'   : {},
                'grid'          : {'visible' : True, 'alpha' : 0.6 },
                'legend'        :{
                    'bbox_to_anchor'    : (1, 1),
                    'loc'               : 'upper right',
                    'ncol'              : 1
                }
            })           

        plt.savefig("%s-lines-verticalstack.png"%(self.objid), **ISMgasPlot['savefig'])   
        plt.close()   
        
    def plotContinuumFit(self, xlim=[-2000,2000], ylim=[0,2]):
        fooDict               = load_pickle(fileName=f"{self.objid}-continuumfit.pkls")
        wav                   = fooDict['wav']
        continuum_fit_poly    = fooDict['continuum_fit_poly']
        final_profile         = fooDict['final_profile']
        final_error           = fooDict['final_error']

        plt.figure(figsize = [12,7])
        plt.plot(
            wav, continuum_fit_poly,
            alpha       = 0.8,
            linewidth   = 3,
            linestyle   = '--',
            label       = 'fit-continuum',
            color       = self.color_pal[16]
        )

        plt.plot(
            wav, final_profile,
            alpha       = 0.8,
            linewidth   = 3,
            label       = 'data',
            color       = self.color_pal[16]
        )
        beautifyPlot({  

            'xlabel'  : {'xlabel': 'Velocity(km/s)','fontsize': 20},
            'ylabel'  : {'ylabel': 'Flux', 'fontsize' : 20},

            'xlim'    : {'left': xlim[0],'right': xlim[1]},
            'ylim'    : {'bottom': ylim[0],'top': ylim[1]},

            'xticks':{
                'ticks'       : np.arange(xlim[0],xlim[1],200),
                'fontsize'    : 15,
                'rotation'    : 90
            },
            'yticks':{
                'ticks'       : np.arange(ylim[0],ylim[1],0.5),
                'fontsize'    : 15
            },
            
            'tightlayout'   : {},
            'legend'        : {},
            'grid'          : {'visible' : True, 'alpha' : 0.6 }

        })
        plt.savefig("%s-fit-continuum.png"%(self.objid), **ISMgasPlot['savefig'])
        plt.close()
        
    def plotCombinedLines(self, xlim=[-2000,2000], ylim=[0,2]):
        fooDict = load_pickle(fileName = f"{self.objid}-combined.pkls")
    
        wav             = fooDict['source_wav']
        final_profile   = fooDict['source_flux']
        final_error     = fooDict['source_error']
        state           = fooDict['state']
        
        plt.figure(figsize = [12,7])
        plotWithError(
            x       = wav,
            y       = final_profile,
            yerr    = final_error
        )
        
        beautifyPlot({  

            'xlabel'  : {'xlabel': 'Velocity(km/s)','fontsize': 20},
            'ylabel'  : {'ylabel': 'Flux', 'fontsize' : 20},

            'xlim'    : {'left': xlim[0],'right': xlim[1]},
            'ylim'    : {'bottom': ylim[0],'top': ylim[1]},

            'xticks':{
                'ticks'       : np.arange(xlim[0],xlim[1],200),
                'fontsize'    : 15,
                'rotation'    : 90
            },
            'yticks':{
                'ticks'       : np.arange(ylim[0],ylim[1],0.5),
                'fontsize'    : 15
            },
            
            'tightlayout'   : {},
            'legend'        : {},
            'grid'          : {'visible' : True, 'alpha' : 0.6 }

        })
        
        fooPlotDict = {
            'lowion'    : {'title': "%s - Absorption profile - Low ionization lines "%(self.objid) , 'saveFile': "%s-lowions-velocityprofile.png"%(self.objid)} ,
            'highion'   : {'title': "%s - Absorption profile - High ionization lines "%(self.objid) , 'saveFile': "%s-highions-velocityprofile.png"%(self.objid)},
            'opthin'    : {'title': "%s - Optically thin lines "%(self.objid), 'saveFile': "%s-opthin-velocityprofile.png"%(self.objid)},
            'stellar'   : {'title': "%s - Absorption profile - Stellar lines "%(self.objid), 'saveFile': "%s-stellar-velocityprofile.png"%(self.objid)},
            'fineem'    : {'title': "%s - Fine structure emission lines "%(self.objid), 'saveFile': "%s-fineem-velocityprofile.png"%(self.objid)},
            'nebem'     : {'title': "%s - Nebular emission lines "%(self.objid), 'saveFile': "%s-nebem-velocityprofile.png"%(self.objid)}
        }
        plt.title(fooPlotDict[state]['title'],fontsize = 20)
        plt.savefig(fooPlotDict[state]['saveFile'], **ISMgasPlot['savefig'])
        plt.close()

    def plotSmoothedSpectra(
            self,
            Nsm,
            display               = "python",
            xlim                  = [None,None],
            ylim                  = [None,None],
            printLinesUsed        = False,
            plotDeflectorLines    = False,
            plotStellar           = False 
        ):
        '''
        Plots weighted average spectra after smooted by Nsm sized kernel
        This can be displayed on a simple python matplotlib.


        - Nsm = number of elements to average over
        - display = 'python' or 'js' - displays the spectra using matplotlib or plotly
        - linelists_highz = ISM lines
        - linelists_stellar = Stellar lines
        - linelists_general = rough lines spanning the entire spectral range.

        After plotting, returns the wavlength,flux and sigma arrays used for plotting
        '''
        
        wave, spec, ivar    = self.fetchData()
        wave_rest           = wave/(1+self.zs)

        f_wtd, ww_wtd, sigma_wtd = returnWeightedArray(
            Nsm          = Nsm,
            spec         = spec,
            ivar         = ivar,
            wave_rest    = wave_rest
        )
        
        save_as_pickle({'Nsm':Nsm, 'flux':f_wtd, 'wav': ww_wtd, 'sigma':sigma_wtd}, "%s-smoothedSpectra.pkls"%(self.objid))
    
        if(display=="python"):
            ''' (default) uses matplotlib '''
            plt.figure(figsize=[15,5])
            font_size=15

            linelist_keys   = linelist_highz.keys()
            if(printLinesUsed):
                print(linelist_keys)

            plt.plot(
                ww_wtd,
                f_wtd,
                alpha       = 0.8,
                linewidth   = 2,
                color       = self.color_pal[11]
            )

            if(plotDeflectorLines):
                colorDef = iter(cm.rainbow(np.linspace(0, 1, 5)))
                for k in self.zdef:
                    colorDefChosen = next(colorDef)
                    for i in linelist_keys:

                        if(
                            linelist_highz[i].get("plot") == True and
                            linelist_highz[i].get("emission") == False  and
                            linelist_highz[i].get("lowion") == True or
                            linelist_highz[i].get("highion") == True
                        ):

                            plt.vlines(
                                x           = linelist_highz[i]["lambda"]*(1+k)/(1+self.zs),
                                ymin        = ylim[1]-0.3,
                                ymax        = ylim[1],
                                color       = colorDefChosen,
                                alpha       = 0.8,
                                linestyle   = '-',
                                linewidth   = 2
                            )

                    plt.plot(
                        0,0,
                        linewidth   = 1,
                        color       = colorDefChosen,
                        alpha       = 0.4,
                        label       = "Inter-abs (z=%.2f)"%(k)
                    )


            if(printLinesUsed):
                print(linelist_keys)

            for i in linelist_keys:
                if(linelist_highz[i].get("lowion") == True and linelist_highz[i].get("plot") == True):
                    plt.vlines(
                        x           = linelist_highz[i]["lambda"],
                        ymin        = ylim[0],
                        ymax        = ylim[1],
                        color       = 'blue',
                        alpha       = 0.8,
                        linestyle   = '--',
                        linewidth   = 2
                    )

                if(linelist_highz[i].get("highion")==True and linelist_highz[i].get("plot")==True):
                    plt.vlines(
                        x           = linelist_highz[i]["lambda"],
                        ymin        = ylim[0],
                        ymax        = ylim[0]+0.3,
                        color       = 'purple',
                        alpha       = 0.8,
                        linestyle   = '--',
                        linewidth   = 2
                    )

                if(plotStellar and linelist_highz[i].get("stellar")==True and linelist_highz[i].get("plot")==True):

                    plt.vlines(
                        x           = linelist_highz[i]["lambda"],
                        ymin        = ylim[0],
                        ymax        = ylim[0]+0.3,
                        color       = 'coral',
                        alpha       = 0.8,
                        linestyle   = '--',
                        linewidth   = 2
                    )

            plt.plot(
                0, 0,
                linewidth   = 5,
                color       = 'blue',
                alpha       = 0.8,
                label       = "ISM-lowion"
            )
            plt.plot(
                0, 0,
                linewidth   = 5,
                color       = 'purple',
                alpha       = 0.8,
                label       = "ISM-highion"
            )

            if(plotStellar):
                plt.plot(
                    0,0,
                    linewidth   = 5,
                    color       = 'coral',
                    alpha       = 0.8,
                    label       = "Stellar"
                )
                
            beautifyPlot({  
                'title'   : {'label': '%s (z=%.5f)'%(self.objid,self.zs), 'fontsize' : font_size},

                'xlabel'  : {'xlabel': "Rest frame wavelength (in $\AA$)",'fontsize': font_size},
                'ylabel'  : {'ylabel': "Normalized Flux", 'fontsize' : font_size},

                'xlim'    : {'left': xlim[0],'right': xlim[1]},
                'ylim'    : {'bottom': ylim[0],'top': ylim[1]},

                'xticks':{
                    'ticks'       : np.arange(xlim[0],xlim[1],100),
                    'fontsize'    : 15,
                },
                'yticks':{
                    'ticks'       : np.arange(ylim[0],ylim[1],0.5),
                    'fontsize'    : 15
                },
                
                'tightlayout'   : {},
                'legend'        : {'loc': 'upper right', 'fontsize': font_size},
                'grid'          : {'visible' : True, 'alpha' : 0.6, 'linestyle':'--', 'linewidth': 1}

            })

            plt.savefig("%s-smooth-spectra.png"%(self.objid), **ISMgasPlot['savefig'])
            plt.close()


### Source:  http://www.ucolick.org/~xavier/DLA/Kingpin/atomic.dat ####
linelist_highz = {
    'H I 1216'    : {'pltLabel':'H-I'       ,"lambda": 1215.6701, "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 0.41640  },
    'Si II 1260'  : {'pltLabel':'SI-II'     ,"lambda": 1260.4221, "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 1.0070   },
    'O I 1302'    : {'pltLabel':'O-I+Si-II' ,"lambda": 1302.1685, "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 0.04887  },
    'Si II 1304'  : {'pltLabel':''          ,"lambda": 1304.3702, "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 0.0940   },
    'C II  1334'  : {'pltLabel':'C-II'      ,"lambda": 1334.5323, "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 0.1278   },
    'Si II 1526'  : {'pltLabel':'Si-II'     ,"lambda": 1526.7066, "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 0.1270   },
    'Fe II 1608'  : {'pltLabel':'Fe-II'     ,"lambda": 1608.4511, "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 0.0580   },
    'Fe II 1611'  : {'pltLabel':''          ,"lambda": 1611.2005, "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 0.00136  },

    'Al II 1670'  : {'pltLabel':'Al-II'     ,"lambda": 1670.7874, "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 1.8800   },
    'Fe II 2344'  : {'pltLabel':'Fe-II'     ,"lambda": 2344.214,  "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 0.1140   },
    'Fe II 2374'  : {'pltLabel':'Fe-II'     ,"lambda": 2374.4612, "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 0.0313   },
    'Fe II 2382'  : {'pltLabel':'Fe-II'     ,"lambda": 2382.765,  "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 0.320    },
    'Fe II 2586'  : {'pltLabel':'Fe-II'     ,"lambda": 2586.6500, "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 0.0691   },
    'Fe II 2600'  : {'pltLabel':'Fe-II'     ,"lambda": 2600.1729, "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 0.2390   },
    'Mg II 2796'  : {'pltLabel':'Mg-II'     ,"lambda": 2796.352,  "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 0.6123   },
    'Mg II 2803'  : {'pltLabel':''          ,"lambda": 2803.531,  "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 0.3054   },
    'Mg I 2852'   : {'pltLabel':'Mg-I'      ,"lambda": 2852.9642, "lowion": True, "plot": True, "emission": False, "comments":'', "fion": 1.8100   },

    'Ni II 1317'  : {'pltLabel':'Ni-II'     ,"lambda": 1317.217 , "opthin": True, "plot": True, "emission": False, "comments":'', "fion": 0.07786  },
    'Ni II 1370'  : {'pltLabel':'Ni-II'     ,"lambda": 1370.131,  "opthin": True, "plot": True, "emission": False, "comments":'', "fion": 0.0769   },
    'Ni II 1709'  : {'pltLabel':'Ni-II'     ,"lambda": 1709.6042, "opthin": True, "plot": True, "emission": False, "comments":'', "fion": 0.0324   },
    'Ni II 1741'  : {'pltLabel':'Ni-II'     ,"lambda": 1741.5531, "opthin": True, "plot": True, "emission": False, "comments":'', "fion": 0.0427   },
    'Ni II 1751'  : {'pltLabel':''          ,"lambda": 1751.915,  "opthin": True, "plot": True, "emission": False, "comments":'', "fion": 0.0277   },
    'Si I 1845'   : {'pltLabel':'Si-I'      ,"lambda": 1845.520,  "opthin": True, "plot": True, "emission": False, "comments":'', "fion": 0.2290   },
    'S II 1250'   : {'pltLabel':'S-II'      ,"lambda": 1250.584,  "opthin": True, "plot": True, "emission": False, "comments":'', "fion": 0.005453 },
    'S II 1253'   : {'pltLabel':'S-II'      ,"lambda": 1253.811,  "opthin": True, "plot": True, "emission": False, "comments":'', "fion": 0.01088  },
    'Si II 1808'  : {'pltLabel':'Si-II'     ,"lambda": 1808.0130, "opthin": True, "plot": True, "emission": False, "comments":'', "fion": 0.002186  },

    'Si IV 1393'    : {'pltLabel': 'Si-IV'    ,"lambda": 1393.755 , "highion": True, "plot": True, "emission": False, "comments":'', "fion": 0.5280  },
    'Si IV 1402'    : {'pltLabel': 'SI-IV'    ,"lambda": 1402.770 , "highion": True, "plot": True, "emission": False, "comments":'', "fion": 0.2620  },
    'C IV  1548'    : {'pltLabel': 'C-IV'     ,"lambda": 1548.195 , "highion": True, "plot": True, "emission": False, "comments":'', "fion": 0.19080 },
    'C IV  1550'    : {'pltLabel': ''         ,"lambda": 1550.770 , "highion": True, "plot": True, "emission": False, "comments":'complicated as there is ISM absorption, and broad stellar P-Cygni absorption + emission', "fion": 0.095220},
    'Al III 1854'   : {'pltLabel': 'Al-III'   ,"lambda": 1854.7164, "highion": True, "plot": True, "emission": False, "comments":'', "fion":0.5390   },
    'Al III 1862'   : {'pltLabel': ''         ,"lambda": 1862.7895, "highion": True, "plot": True, "emission": False, "comments":'', "fion":0.2680   },
}



class GalaxyProperties:
    def __init__(self,**kwargs):
        self.objid            = kwargs.get('objid')
        self.ra               = kwargs.get('ra')            ## ra in degrees
        self.dec              = kwargs.get('dec')           ## dec in degrees
        self.zs               = kwargs.get('zs')            ## redshift of the source galaxy
        self.zsNotes          = kwargs.get('zsNotes')
        self.zdef             = kwargs.get('zdef',[])       ## redshift of the deflector galaxy if known
        self.zdefNotes        = kwargs.get('zdefNotes')
        self.inst             = kwargs.get('inst')          ## Instrument used to observe target e.g ESI/NIRES/etc
        self.inst_pipeline    = kwargs.get('inst_pipeline') ## pipeline used to reduce the data
        self.inst_sigma       = kwargs.get('inst_sigma',1)  ## Instrument resolution sigma in km/s
        self.spec_filename    = kwargs.get('spec_filename') ## filename containing the fits file
        self.mass             = kwargs.get('mass',[])
        self.sfr              = kwargs.get('sfr',[])
        self.survey           = kwargs.get('survey')
        self.merger           = kwargs.get('merger')
        self.mergerNotes      = kwargs.get('mergerNotes')
        self.R                = kwargs.get('R')



import os
import pickle

import numpy as np
import pandas as pd

import scipy.stats as stat
from scipy import interpolate
from scipy.constants import c, pi
c_kms = c*1e-3

import matplotlib.pyplot as plt
import matplotlib.image as mpimg
from matplotlib import gridspec
import matplotlib.patches as patches

from astropy.io import fits
from astropy.visualization import hist 
from astropy.cosmology import FlatLambdaCDM
from astropy.wcs import WCS
from astropy.cosmology import FlatLambdaCDM
from astropy.coordinates import SkyCoord
from astropy import units as u
from astropy.table import Table

#########################
## Astronomy utilities ##
#########################
def convertDegreesToHMS(ra_deg:float ,dec_deg:float)->str:
    '''
    returns ra and dec in hms from degrees using astropy
    '''
    c = SkyCoord(ra=ra_deg*u.degree, dec=dec_deg*u.degree)
    return(c.to_string('hmsdms').replace('h',':').replace('d',':').replace('m',':').replace('s',''))


def save_spectra(wave, flux, error, fileName, folderPrefix = ''):

    t             = Table()
    t['LAMBDA']   = [wave]
    t['SPEC']     = [flux]
    t['ERR']      = [error]
    t['IVAR']     = [1/error**2] ## 1/sigma^^2 = ivar
    
    t.write(folderPrefix+"%s.fits"%(fileName),overwrite=True)

    print("Written file to " + folderPrefix+"%s.fits"%(fileName))

############
### Math ###
############
def interpolateData(x, y, xnew):
    '''
    Assuming y = f(x), returns f(xnew) using scipy's interpolate method.
    '''
    
    fInter  = interpolate.interp1d(x, y)
    return fInter(xnew)


def returnWeightedArray(Nsm, spec, ivar, wave_rest):
    '''
    #### Logic

    $$  wavelength = [\lambda_0,\lambda_1,\lambda_2.......\lambda_n] $$

    $$  flux = [f_0,f_1,f_2,......f_n] $$

    $$  ivar= [iv_0,iv_1,iv_2.......iv_n] $$

    $$ f_{iv} = flux*ivar = [f_0 iv_0 , f_1 iv_1, ... f_n iv_n] $$

    $$ f_{weighted} = Convolve(flux* ivar, kernel size)/Convolve(ivar, kernel size)$$

    $$ standard error  = \sqrt{1/\sum /\sigma_i^2}  = \sqrt{1/1\sum ivar_i} =  \sqrt{1/Convolve(ivar,kernel size)}$$

    $$ \lambda_{weighted} = Convolve(wavlength,kernel size)  $$

    This ensures that the $f_{weighted}$, standard error and the $\lambda_{weighted}$ have the same number of elements.

    #### Input
    - Nsm : Kernel size
    - spec : flux array
    - ivar : invariance array (or 1/weights**2 array)
    - wave_rest : rest frame wavelength

    #### Output
    - weighted average flux
    - weighted averate sigma
    - corrected wavelength

    #### Example :
    ```
    kernel_size= 3
    spec = np.array([1,2,3])
    ivar = np.array([10,100,10])
    wave_rest = np.array([1000,2000,3000])

    returnWeightedArray(kernel_size,spec,ivar,wave_rest)

    [array([2.]), array([2000.]), array([0.09128709])]

    ```
    '''
    return([
        np.convolve(spec*ivar, np.ones(Nsm),mode = 'valid')/np.convolve(ivar, np.ones(Nsm),mode = 'valid'),
        np.convolve(wave_rest, np.ones((Nsm,))/Nsm, mode='valid'),
        1/np.sqrt(np.convolve(ivar, np.ones((Nsm)), mode='valid'))
    ])
    
def wavelengthToVelocity(WLarray, lambda0):
    return(c_kms*((np.asarray(WLarray)-lambda0)/lambda0))
    
def find_nearest(input_array, value:float):
    array   = np.asarray(input_array)
    idx     = (np.abs(array - value)).argmin()
    return array[idx]   

def scipyMuSigma(array):
    mu, sigma = stat.norm.fit(array)
    return(mu, sigma)

def medianMAD(array):
    mu    = np.median(array,axis=0)
    MAD   = np.median(np.abs(array-mu))
    ##Approximately Converting MAD to sigma - https://en.wikipedia.org/wiki/Median_absolute_deviation   
    sigma = MAD/0.675
    
    return(mu, sigma)

def chooseAndPropError(x, xerr, y, yerr, n = 1000):
    '''
    This function propagates the errors by randomly sampling the quantities 
    within the errors provided. Assumes a normal distribution for the xerr and yerr quantities.
    
    Input:
    
    x, xerr: DG fit quantity and its error
    y, yerr: SG fit quantity and its error
        
    '''
    assert ((len(x) == len(y)) and (len(x) == len(xerr)) and (len(y) == len(yerr))), "Ensure that length of arrays are same"
    allValues = []
    allStds = []
    for j in range(n):
        foo_std = []
        for i in range(len(x)):
            chosen_x        = np.random.normal(x[i], xerr[i], 1)
            chosen_y        = np.random.normal(y[i], yerr[i], 1)
            quantity_diff   = chosen_x - chosen_y

            foo_std.append(quantity_diff)
            
        allStds.append(np.std(foo_std))
        allValues.append(np.mean(foo_std))

    allValues = np.array(allValues) 
    allStds = np.array(allStds)
    print("Mean:%d $\pm$ %d, St.Dev: %d $\pm$ %d"%(np.mean(allValues), np.std(allValues),np.mean(allStds),np.std(allStds)))
    return([np.mean(allValues), np.std(allValues),np.mean(allStds),np.std(allStds)])

###############################
### Uncategorized functions ###
###############################

def save_as_pickle(arrayToSave,fileName:str):
    pickle.dump(arrayToSave, open( fileName, "wb" ) )
    print("file saved to: "+fileName)

def load_pickle(fileName:str):
    return(pickle.load( open(fileName, "rb" ) ))

def image_to_HTML(path:str,width=600):
    return '<img src="'+ path + '" width="'+ str(width)+'" >'

def makeDirectory(folder):
    try: 
        os.makedirs(folder)
        
    except OSError:
        if not os.path.isdir(folder):
            raise

###############################
##### Operations on Images ####
###############################

def image_array(filename:str):
    return(mpimg.imread(filename))

def display_image(filename:str):
    return(plt.imshow(mpimg.imread(filename)))


###########################
########## Plots ###############
###########################

def beautifyPlot(kwargs):
    '''
    Run at end of matplotlib plotting routine. Thanks to Patrick Wells (UCD) for suggesting a clean way to implement this 
    Example:
    ```
    x   = np.arange(0,100,1)
    y   = x**2

    plt.plot(x,y, label= 'y=f(x)')

    beautifyPlot({
        'title'         : {'label': '$x^2$ v/s x'},
        'xlim'          : {'left': 0, 'right':100},
        'ylim'          : {'bottom': 0, 'top':10000},
        'legend'        : {'bbox_to_anchor' : (1, 1), 'fontsize': 15},
        'tightlayout'   : {}

    })

    ```
    
    '''
    pltFunctions = {
        "title"         : plt.title,
        "xlabel"        : plt.xlabel,
        "ylabel"        : plt.ylabel,
        "xlim"          : plt.xlim,
        "ylim"          : plt.ylim,
        "xticks"        : plt.xticks,
        "yticks"        : plt.yticks,
        "tightlayout"   : plt.tight_layout,
        "legend"        : plt.legend,
        "grid"          : plt.grid,
        "savefig"       : plt.savefig,
        "suptitle"      : plt.suptitle
    }

    for key, arguments in kwargs.items():
        function = pltFunctions.get(key, False)
        if function:
            function(**arguments)
        else:
            print(f"Function {key} is not supported!")


def plotWithError(x, y, yerr, sigmaLimit = 1, label = 'data', **kwargs):
    "By default plots x,y and 1 sigma error region"
    
    plt.plot(
        x,
        y,
        alpha       = 0.8,
        linewidth   = kwargs.get('linewidth', 3),
        label       = label,
        color       = color_pal_kvgc['pub1'][16]
    )

    plt.fill_between(
        x,
        y - yerr * sigmaLimit,
        y + yerr * sigmaLimit,
        alpha       = 0.5,
        facecolor   = color_pal_kvgc['pub1'][7]
    )


def plotHistogram(
        arrayToPlot,
        arrayName       = '',
        bins            = 'scott',
        method          = 2,
        best_fit_plot   = True,
        plotting        = True
    ):
    """
    Returns 
    method = 1 -- (mu,sigma) using scipy gaussnorm -- great for gaussian like distributions
    method = 2 -- (Median,MAD) using median and MAD -- great for non-uniform distributions
    for a given array
    """
    arrayToPlot = np.array(arrayToPlot)

    if(method==1):
        ## Uses scipy norm fit -- works great for gaussian like distribution
        
        mu, sigma     = scipyMuSigma(arrayToPlot)
        if(plotting):
            _, bins, _ = hist(
                arrayToPlot,
                color       = 'black',
                linewidth   = 6,
                alpha       = 0.8,
                bins        = bins,
                histtype    = 'step',
                density     = True
            )

            best_fit_line = stat.norm.pdf(bins, mu, sigma)
            plt.plot(bins, best_fit_line)
            plt.axvline(mu,linestyle='-',linewidth=4,color='purple',alpha=0.7)
            plt.axvline(mu+sigma,linestyle='--',linewidth=2.5,color='green',alpha=0.5)
            plt.axvline(mu-sigma,linestyle='--',linewidth=2.5,color='green',alpha=0.5)
            plt.title("$\mu_{fit}$=%.3f $\sigma_{fit}=$%.3f"%(mu,sigma),fontsize=15)

            plt.xticks(fontsize=15,rotation=90)
            plt.yticks(fontsize=15)
            plt.ylabel("Counts (Normalized)",fontsize=20)
            plt.xlabel(arrayName,fontsize=20)
            plt.grid('on',alpha=0.6)

    elif(method==2):

        ##### MAD - Median absolute deviation-- works great for non-gaussian distribution ####
        mu, sigma     = medianMAD(arrayToPlot)

        if(plotting):
            _, bins, _ = hist(
                arrayToPlot,
                color       = 'black',
                linewidth   = 6,
                alpha       = 0.8,
                bins        = bins,
                histtype    = 'step',
                density     = True
            )

            if(best_fit_plot):
                best_fit_line = stat.norm.pdf(bins, mu, sigma)
                plt.plot(bins, best_fit_line)
                plt.axvline(mu,linestyle='-',linewidth=4, color='purple',alpha=0.7)
                plt.axvline(mu+sigma,linestyle='--',linewidth=2.5,color='green',alpha=0.5)
                plt.axvline(mu-sigma,linestyle='--',linewidth=2.5,color='green',alpha=0.5)
                plt.title("Median=%.1f 1$\sigma$=%.2f"%(mu,sigma),fontsize=15)

            plt.xticks(fontsize=15,rotation=90)
            plt.yticks(fontsize=15)
            plt.ylabel("Counts (Normalized)",fontsize=20)
            plt.xlabel(arrayName,fontsize=20)
            plt.grid('on',alpha=0.6)

    return(mu,sigma)

def drawRectange(x,
                 y,
                 deltax,
                 deltay,
                 linewidth=1,
                 edgecolor='r',
                 facecolor='none'):
    rect = patches.Rectangle((x, y), deltax,deltay,linewidth=linewidth, edgecolor=edgecolor, facecolor=facecolor)
    return(rect)
###########################


from ISMgas.kcwi.kcwiFunctions import *
from ISMgas.visualization.fits import ScaleImage

class kcwiReduxMontage:
    def __init__(self, galaxy, filenames, pixscale=0.3, **kwargs):
        self.ra = galaxy.ra
        self.dec = galaxy.dec
        self.objid = galaxy.objid
        self.filenames = filenames
        self.pixscale = pixscale ## arcseconds
        self.dataCube  = None
        
              
    def combineAndDrizzle(self, drizzleFactor=0.7):
        ### Write a DECaLS like header
        template = f"""
SIMPLE  =                    T / file does conform to FITS standard
BITPIX  =                  -32 / number of bits per data pixel
NAXIS   =                    2 / number of data axes
NAXIS1  =                  140 / length of data axis 1
NAXIS2  =                  140 / length of data axis 2
EXTEND  =                    T / FITS dataset may contain extensions
COMMENT   FITS (Flexible Image Transport System) format is defined in 'Astronomy
COMMENT   and Astrophysics', volume 376, page 359; bibcode: 2001A&A...376..359H
SURVEY  = 'LegacySurvey'
VERSION = 'DR9     '
IMAGETYP= 'IMAGE   '           / None
CTYPE1  = 'RA---TAN'           / TANgent plane
CTYPE2  = 'DEC--TAN'           / TANgent plane
CRVAL1  =              {self.ra} / Reference RA
CRVAL2  =             {self.dec} / Reference Dec
CRPIX1  =                 70.5 / Reference x
CRPIX2  =                 70.5 / Reference y
CD1_1   = -{self.pixscale/3600} / CD matrix
CD1_2   =                   0. / CD matrix
CD2_1   =                   0. / CD matrix
CD2_2   = {self.pixscale/3600} / CD matrix
END
"""
        f = open(f"{self.objid}.hdr", 'w+')
        f.write(template)
        f.close()        
        
        dd = kcwiAnalysis(
            filename = self.filenames,
        )

        fits.writeto(
            filename = self.objid + ".fits",
            data = dd.dataCube,
            header = fits.getheader(self.filenames[0]),
            overwrite= True
        )          

        ## Project datacube using montage
        cmd = f"mProjectCube -X -z 0.7 {self.objid}.fits  {self.objid}_drizzle.fits  {self.objid}.hdr"
        runCMD(cmd)
        print(f"Drizzled file: {self.objid}_drizzle.fits")
        
        ## Show user the drizzled datacube
        self.dataCube = fits.getdata(f"{self.objid}_drizzle.fits")
        dd = kcwiAnalysis(
            objid = 'test',
            filename = [f"{self.objid}_drizzle.fits"],
        )
        print(f"Shape: {np.shape(dd.dataCube)}")

        ScaleImage(dd.dataCubeMean).plot()
        

def padAndAlign(cubes, newOutputShape, centroids=[]):
    
    results = []
    alignStack = []
        
    for i in range(len(cubes)):
        
        newCube = np.zeros(newOutputShape)
        newCube[:,0:cubes[i].dataCube.shape[1], 0:cubes[i].dataCube.shape[2]] = cubes[i].dataCube

        
        if(len(centroids)==0):
            results.append(newCube)
            alignStack.append(np.mean(newCube[500:-500, :, :], axis=0))
            
        else:
            newAlignCube = np.roll(
                newCube, 
                [0, -(centroids[i][0] - centroids[0][0]), -(centroids[i][1]- centroids[0][1])],
                axis=(0,2,1)
            )
            results.append(newAlignCube)
            alignStack.append(np.mean(newAlignCube[500:-500, :, :], axis=0))
        
    fits.writeto(
        filename = 'align.fits',
        data = np.array(alignStack),
        overwrite= True
    )
    print("Use align.fits to manually align the datacubes")
    return(np.mean(results,axis=0))
    